--[[
Rewritten from Incursion Guild Alts, originally by Zorlan.  See there for the
full tortured history.

]]

local nametag, addon = ...

local altrank = 'alt'   -- must be lowercase.  FIXME  needs to allow for l10n

addon.defaults = {
	profile = {
		enable = true,
		use_alt_rank = false,
		alt_pattern = "[Aa][lL][tT][ ]?%(<NAME>%)",
		tag_channels = "raid",
		do_rank_tags = false,
		do_rank_tags_officer = false,
		debug = false,
		hcd_debug = false,
		debugging_frame = 1,
		hcd_debugging_frame = 2,
		max_chatframe_num = 7,
	},
	global = {
		guilds = { },
	},
}


local pattern_editing_safe
addon.options = {
	name = "Ouro Guild Alts",
	type = 'group',
	childGroups = 'tab',
	handler = addon,    -- functions listed as strings called as addon:func
	get = "GetOption",
	set = "SetOption",
	args = {
		general = {
			name = "General",
			desc = "General options",
			type = 'group',
			order = 10,
			args = {
				version = {
					--name = filled in during OnInit
					type = 'description',
					fontSize = "large",
					cmdHidden = true,
					width = 'full',
					order = 1,
				},
				enable = {
					name = "Enable",
					desc = "Use this addon",
					type = 'toggle',
					arg  = "ToggleEnable",
					order = 5,
				},
				rescan = {
					name = "Rescan Guild Roster",
					desc = "Reloads the list of alts from the guild roster",
					type = 'execute',
					func = "LoadGuild",
					order = 10,
				},
				--[[spacer = {
					name = '',
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = 14,
				},]]
				use_alt_rank = {
					name = "Check Ranks",
					--desc = -- filled in below,
					type = 'toggle',
					width = 'double',
					order = 15,
				},
				instructions = {
					name = [[Bracketed name replacement will take place in guild, officer, party, and raid chat.  All settings are per-profile except for the abbreviations used in rank tags, which are stored for each guild and configurable by all characters.]],
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = -1,
				},
			},
		},
		ranktags = {
			name = "Rank Tags",
			desc = "How and when to use them",
			type = 'group',
			order = 30,
			get = false, set = false,
			args = {
				-- rank-tag-related stuff filled in below
			},
		},
		advanced = {
			name = "Advanced",
			desc = "Unusual options",
			type = 'group',
			order = 50,
			args = {
				debug = {
					name = "Toggle debugging",
					desc = "This is a metric assload of text output, especially if HCD is active",
					type = 'toggle',
					arg  = "HookChat",
					order = 55,
				},
				hcd_debug = {
					name = "Toggle HCD",
					desc = "You really want to put this in its own window. |cffff0000REQUIRES /reload AFTER ENABLING.|r",
					type = 'toggle',
					order = 56,
				},
				print_chatframes = {
					name = "Print Chatframe Numbers",
					desc = "Print each chat window number in its frame, for easy reference in the next slider options",
					type = 'execute',
					--func = print_chatframes filled in below
					order = 57,
				},
				debugging_frame = {
					name = "Debugging Output Chatframe",
					desc = "Which chat window to use for normal debug output",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					order = 58,
				},
				hcd_debugging_frame = {
					name = "HCD Debugging Output Chatframe",
					desc = "Which chat window to use for HCD output. |cffff0000REQUIRES /reload AFTER CHANGING.|r",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					order = 59,
				},
				max_chatframe_num = {
					name = "Max Chatframe",
					desc = "Highest chatframe number to hook",
					type = 'range',
					min  = 1,
					max  = NUM_CHAT_WINDOWS,
					step = 1,
					arg  = "HookChat",
					order = 60,
				},
				-- If they screw up the input field, validation error messages are lost
				-- http://www.wowace.com/addons/ace3/tickets/2
				alt_pattern_warning = {
					name = [[|cff5500ffATTENTION!|r  If you plan on changing the Pattern field here, you must use the "/oga" command to display this menu instead of Blizzard's "Escape -> Interface -> Addons" menu.  This is due to a bug not under my control.  When it is fixed, this annoying text will be removed.]],
					type = 'description',
					hidden = function() return pattern_editing_safe end,
					cmdHidden = true,
					order = 64,
				},
				alt_pattern = {
					name = "Pattern",
					desc = "Lua pattern to detect and extract |cff00ff00<NAME>|r out of the player and officer note fields (see HOWTO.txt for more).",
					type = 'input',
					hidden = function() return not pattern_editing_safe end,
					validate = "UpdatePattern",
					order = 65,
				},
			},
		},
		--profiles =   filled in OnInit
	}
}


-----------------------------------------------------------------------------
-- other locals
local guildname, guildkey, isofficer = 'NONE', 'NONE', false
local flib = LibStub("LibFarmbuyer", true)
local lua_pattern = "PLACEHOLDER THAT WILL NEVER MATCH BUT WILL ALSO BE GARBAGE COLLECTED"
local state = setmetatable({
	me = UnitName("player"),
	do_rank_tags = false,
	orig = DEFAULT_CHAT_FRAME.AddMessage, -- on the assumption that all AddMessage's are the same
	channel_match_pattern = false,
	extract_name_and_target = false,
	author_debug = not not (flib and flib.author_debug),   -- must be boolean
	hcd_debugging_frame = false,
	debugging = function() return addon.db.profile.debug end,
	tag_channels = {},
}, {__index = _G})

-- First argument may optionally be a specific chatframe to use, or a boolean
-- overriding the current debug setting.
state.dprint = function(...)
	local t = "<D>"
	local frame
	local do_it = addon.db.profile.debug
	local first = ...
	if type(first) == "table" and first.AddMessage then
		frame, first = first, 2
	elseif type(first) == "boolean" then
		frame, first = addon.db.profile.debugging_frame, 2
		do_it = do_it or first
	else
		frame, first = addon.db.profile.debugging_frame, 1
	end
	if not do_it then return end
	if type(frame) ~= "table" then
		frame = _G["ChatFrame"..frame]
	end
	for i = first, select('#',...) do
		local x = select(i,...)
		t = t .. " " .. tostring(x)   -- take that, garbage collector...
	end
	state.orig(frame,t,1,0,0)  -- specifically not tailcalling
end

local function print_chatframes()
	for i = 1, NUM_CHAT_WINDOWS do
		local cf = _G["ChatFrame"..i]
		if not cf then break end
		state.orig(cf,"This is frame number " .. i .. ".",1,1,1)
	end
end
addon.options.args.advanced.args.print_chatframes.func = print_chatframes
addon.options.args.general.args.use_alt_rank.desc =
	"In addition to notes, always include members whose rank includes the "..
	"word '"..altrank.."' in some form, and use the entire note field as "..
	"their main's name."


-- Deal with Prat's various possibilities.  If it's not in use, then do our
-- own chat hooks.  If their alt module is fired up, then add our markup pattern
-- (none of its defaults will match).  Otherise hook into their message format
-- function.
--
-- Returns:
--   "normal" if we should keep doing our thing
--   "bail" if we should leave maps in place but not change chat
--   "cleaup" if we should bail and also clear our own memory
--
--local DO = 0
local function handle_prat()
	if not Prat then return "normal" end
	local mod = Prat.Addon:GetModule("AltNames", --[[silent=]]true)
	if not mod then return "normal" end
	if mod:IsEnabled() then
		local addalt = mod.addAlt
		local pattern = "(%S+)%s+" .. lua_pattern
		function mod:addAlt(str)
			local alt,main = str:match(pattern)
			if main then return addalt(self, alt..' '..main)
			else return addalt(self,str)  -- not one of ours
			end
		end
		return "cleanup"
	end
	local function hook(...)
		--if DO == 2 then return end
		local e,m = ...
		local player = m.PLAYERLINK
		--print(e, "player is", player, "escaped is", player:gsub('\124', '\124\124'))
		if player ~= state.me or true then
			--print('replacing', player)
			local mainname = addon.alts_to_main[player]
			--print('main is', mainname)
			if mainname and mainname ~= "" then
				m.PLAYER = ("%s (%s)"):format(player, mainname)
				--print('m.player now', m.PLAYER)
				--DO = DO + 1
			end
		end
	end
	Prat.RegisterChatEvent(nametag, "Prat_FrameMessage", hook)
	--[==[
	Prat.RegisterChatEvent("OuroGuildAlts", "Prat_PreAddMessage", function(...)
		if DO == 2 then return end
		local e,m,_,_,out = ...
		print(e, out:gsub('\124', '\124\124'))
				if DO == 1 then DO = DO + 1 end
	end)
	]==]
	return "bail"
end


-----------------------------------------------------------------------------
addon = LibStub("AceAddon-3.0"):NewAddon(addon, nametag,
                "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0")

function addon:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("OuroGuildAltsDB", self.defaults, --[[Default=]]true)

	if not next(self.db.global.guilds) then
		self:load_Ouroboros_defaults()
	end
	self.load_Ouroboros_defaults = nil

	for c in self.db.profile.tag_channels:gmatch("[^,]+") do
		state.tag_channels[c] = true
		state.tag_channels[#state.tag_channels+1] = c
	end

	local AceDBOptions = LibStub("AceDBOptions-3.0", true)
	if AceDBOptions then
		self.options.args.profiles = AceDBOptions:GetOptionsTable(self.db)
		self.options.args.profiles.order = 200
	end

	self.options.args.general.args.version.name = 
		"|cff30adffVersion " .. (GetAddOnMetadata(nametag, "Version") or "?") .. "|r"
	LibStub("AceConfig-3.0"):RegisterOptionsTable(nametag, self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(nametag, "Ouro Guild Alts")
	self.optionsFrame.okay = function() pattern_editing_safe = false end
	self.optionsFrame.refresh = self.optionsFrame.okay
	self.optionsFrame.cancel = self.optionsFrame.okay
	self:RegisterChatCommand("oga", "OnChatCommand")

	self:SetEnabledState(self.db.profile.enable)
	self.OnInitialize = nil
end


--[[
function addon:FriendsFrame_Pause()
	FriendsFrame:UnregisterEvent("WHO_LIST_UPDATE")
end
function addon:FriendsFrame_Resume()
	FriendsFrame:RegisterEvent("WHO_LIST_UPDATE")
end
]]


function addon:OnEnable()
	self:RegisterEvent("PLAYER_LOGOUT")
	if not self.db.profile.enable then
		return self:Disable()
	end

	if self.prat == "cleanup" then
		return self:Print("ERROR: After turning off Prat modules, you will need to relog or /reload to get this addon working.")
	end

	-- Blizzard's code dispatches a GuildRoster() call 15 seconds after load.
	-- Sometimes these events fire before the data is actually present, leading
	-- to screwups during the scan.  Waiting a little while if possible is better.
	local maybe = LibStub("AceTimer-3.0",true)  -- only used once, not worth shipping
	if maybe then
		maybe.ScheduleTimer(self, function()
				self:RegisterEvent("GUILD_ROSTER_SHOW", "LoadGuild")
				self:RegisterEvent("GUILD_ROSTER_UPDATE", "LoadGuild")
			end, 2)
	else
		-- enh, register now and hope that the lagloop in LoadGuild is enough
		self:RegisterEvent("GUILD_ROSTER_SHOW", "LoadGuild")
		self:RegisterEvent("GUILD_ROSTER_UPDATE", "LoadGuild")
	end
end

function addon:OnDisable()
	self:UnhookChat()
end

function addon:ToggleEnable()
	if self.db.profile.enable then
		self:Enable()
		GuildRoster()
	else
		self:Disable()
	end
end

function addon:PLAYER_LOGOUT()
	--self.db.profile.debug = nil
	self.db.global.guilds.NONE = nil
	for name,gtab in pairs(self.db.global.guilds) do
		if not next(gtab.abbr) then
			table.wipe(self.db.global.guilds[name])
			self.db.global.guilds[name] = nil
		end
	end
	self.db.profile.tag_channels = table.concat(state.tag_channels,',')
end


do
	local optgroup = addon.options.args.ranktags.args

	optgroup.do_rank_tags = {
		name = "Use rank tags",
		desc = "Add short rank-based tags to names",
		type = 'toggle',
		get = "GetOption", set = "SetOption",
		arg  = "UpdateRankTagging",
		order = 15,
	}

	optgroup.rank_tag_control = {
		name = "Rank tag details",
		type = 'group',
		inline = true,
		get = false, set = false,
		disabled = function(i) return not i.handler.db.profile.do_rank_tags end,
		arg  = "UpdateRankTagging",
		order = 20,
		args = {},
	}
	local RTdisabled = optgroup.rank_tag_control.disabled

	optgroup.warn = {
		name = [[Due to Blizzard API limitations, a rank tag for a guildmate will only start taking effect when that guildmate is nearby and (usually) grouped with you.  To conserve memory and CPU cycles, Ouro Guild Alts intentionally makes no special efforts to work around these limitations.]],
		type = 'description',
		cmdHidden = true,
		order = 30,
	}

	optgroup = optgroup.rank_tag_control.args

	optgroup.chanlist = {
		name = "Channels",
		desc = "Add rank tags in these channels",
		type = 'multiselect',
		dialogControl = "Dropdown",
		order = 5,
		-- these need to be of function type rather than string keys of members
		values = function(info) return addon:ChannelList() end,
		get = function(info,x) return addon:GetChannel(x) end,
		set = function(info,val) return addon:SetChannel(val) end,
	}

	optgroup.do_rank_tags_officer = {
		name = "Tags only when officer",
		desc = "Use rank tags only when you are a guild officer ('officer' defined as being able to /ginvite)",
		type = 'toggle',
		get = "GetOption", set = "SetOption",
		arg  = "UpdateRankTagging",
		order = 10,
	}
	optgroup.spacer1 = {
		name = '',
		type = 'description',
		cmdHidden = true,
		width = 'full',
		order = 20,
	}

	local guildlist_v                -- known guild keys
	local workingguild = guildkey    -- current dropdown selection
	local function no_g_selection()
		return workingguild == 'NONE' and true or (not workingguild)
	end

	optgroup.guildlist = {
		name = "Guild",
		desc = "Which guild to configure",
		type = 'select',
		order = 21,
		width = 'double',
		values = function(info)
			if not guildlist_v then
				-- Build a list of all previously configured guild-realm combos, plus
				-- the current guild whether it's in the list or not.  (This is the only
				-- way we allow of getting new combos potentially added to the list.)
				guildlist_v = { [guildkey] = guildkey }
				for gk in pairs(addon.db.global.guilds) do
					guildlist_v[gk] = gk
				end
			end
			return guildlist_v
		end,
		get = function(info) return workingguild == 'NONE' and guildkey or workingguild end,
		set = function (info, key) workingguild = key end,
	}

	local workingrank
	local function no_r_selection() return not workingrank end
	optgroup.ranklist = {
		name = "Rank",
		desc = "List of ranks for the guild to the left",
		type = 'select',
		order = 22,
		disabled = function(i) return RTdisabled(i) or no_g_selection() end,
		values = function(info)
			if RTdisabled(info) then return {} end   -- meh
			local g = addon.db.global.guilds
			if not g[workingguild] then
				-- configuring our current guild for the first time
				g[workingguild] = { abbr = {}, ranks = {} }
			end
			g = g[workingguild]
			if (not no_g_selection()) and workingguild == guildkey then
				-- it's our current guild, we can refresh the ranks
				g.ranks = {}
				for i=1, GuildControlGetNumRanks() do
					g.ranks[i-1] = GuildControlGetRankName(i)
				end
			else
				-- must use stored ranks for other guilds
			end
			return g.ranks
		end,
		get = function(info) return workingrank or '' end,
		set = function (info, key) workingrank = key end,
	}

	optgroup.abbredit = {
		name = "Abbreviation",
		desc = "What to append to names for the selected rank",
		type = 'input',
		order = 23,
		width = 'half',
		disabled = function(i) return RTdisabled(i) or no_r_selection() end,
		get = function(info)
			if no_g_selection() or no_r_selection() then return '' end
			local g,r = addon.db.global.guilds[workingguild], tonumber(workingrank) or 0
			return g.abbr[r]
		end,
		set = function (info, val)
			if no_g_selection() or no_r_selection() then return end
			val = tostring(val)
			if not val then return end
			local g,r = addon.db.global.guilds[workingguild], tonumber(workingrank)
			g.abbr[r] = val ~= '' and val or nil
		end,
	}

	optgroup.spacer2 = {
		name = '',
		type = 'description',
		cmdHidden = true,
		width = 'full',
		order = 40,
	}
	optgroup.clear = {
		name = "Clear Guild",
		desc = "Clear settings for selected guild",
		type = 'execute',
		order = 42,
		disabled = function(i) return RTdisabled(i) or no_g_selection() end,
		func = function(info)
			-- Yes, they need to be wiped before being nil'd
			-- http://forums.wowace.com/showthread.php?p=307195#post307195
			table.wipe(addon.db.global.guilds[workingguild])
			addon.db.global.guilds[workingguild] = nil
			-- Update the UI as well
			guildlist_v[workingguild] = nil
			workingrank = nil
			info.handler:Printf([[Configuration for %s%s|r removed.  If this was your current guild, you will need to relog, /reload, or click "%sRescan Guild Roster|r" before the change takes effect.]],
				GREEN_FONT_COLOR_CODE,
				workingguild,
				NORMAL_FONT_COLOR_CODE)
			workingguild = 'NONE'
		end,
	}

	function addon:PrepareRankTagging()
		workingguild = guildkey

		self.PrepareRankTagging = nil
	end

	local mt = {
		__index =
			function(t,name)
				state.dprint("rank_lookup starting", name)
				local gn,gr,rank = GetGuildInfo(name)  -- returns 0-based rank
				state.dprint(gn, gr, rank, "and stored guildname is", guildname)
				if gn ~= guildname then return nil end
				rank = addon.rank_letter_map[rank]
				if not rank then return nil end
				t[name] = rank
				state.dprint("rank_lookup returning", name, rank)
				return rank
			end
	}

	function addon:UpdateRankTagging()
		state.do_rank_tags = self.db.profile.do_rank_tags
		if self.db.profile.do_rank_tags_officer then
			state.do_rank_tags = state.do_rank_tags and isofficer
		end
		if state.do_rank_tags and not self.rank_lookup then
			self.rank_lookup = setmetatable({}, mt)
			local rlm = self.db.global.guilds[guildkey]
			rlm = rlm and rlm.abbr
			-- If it doesn't exist, force replacements to silently fail by
			-- using an empty-ish table.  The metatable is empty enough for
			-- this purpose and saves creating a dummy table.
			self.rank_letter_map = rlm or mt
		end
	end
end


do
	local function disp(txt)
		return GREEN_FONT_COLOR_CODE .. txt .. FONT_COLOR_CODE_CLOSE
	end

	function addon:ChannelList()
		-- cribbed from Acheron
		local display_channel_list = {
			["say"] = disp"say",
			["party"] = disp"party",
			["raid"] = disp"raid",
			["guild"] = disp"guild",
			["officer"] = disp"officer",
		}

		local numbered = {}
		for _,chan in ipairs{EnumerateServerChannels()} do
			numbered[chan] = true
		end

		local joined = {GetChannelList()}
		for i=1, #joined, 2 do
			-- don't include the public numbered channels
			if not numbered[ joined[i+1] ] then
				display_channel_list[ tostring(joined[i]) ] = disp(joined[i+1])
			end
		end

		return display_channel_list
	end

	function addon:GetChannel (name)
		return state.tag_channels[name]
	end

	function addon:SetChannel (name)
		local was_on = not not state.tag_channels[name]
		if was_on then
			state.tag_channels[name] = nil
			for i,v in ipairs(state.tag_channels) do
				if v == name then
					table.remove(state.tag_channels,i)
					break
				end
			end
		else
			state.tag_channels[name] = true
			state.tag_channels[#state.tag_channels+1] = name
		end
	end
end


-- Thanks to jerry for the nifty arg idea.
function addon:SetOption (info, value)
	local name = info[#info]
	self.db.profile[name] = value
	local arg = info.arg
	if arg then self[arg](self) end
end

function addon:GetOption (info)
	local name = info[#info]
	return self.db.profile[name]
end


do
	-- Try to compensate for AceGUI ticket #2.
	local function err (from_menu, txt)
		if from_menu then
			return txt
		else
			addon:Print("Error: ", txt)
		end
	end
	function addon:UpdatePattern (from_menu, pattern)
		pattern = tostring(pattern)
		if not pattern then return err(from_menu,"The pattern must be an actual string...") end

		local pat, count = pattern:gsub('<NAME>$', '(.-)$')
		if count == 0 then
			pat, count = pattern:gsub('<NAME>', '(.-)')
		end
		if count ~= 1 then
			return err(from_menu,
				([[The string "<NAME>" must appear exactly once in the pattern "%s", currently %s.  Close and reopen the window to try again.]]):format(pattern,tostring(count)))
		end
		lua_pattern = pat
		state.dprint("UpdatePattern arrives at:", lua_pattern)
		return true
	end
end


function addon:OnChatCommand (input)
	pattern_editing_safe = true
	if not input or input:trim() == "" then
		-- running this before the guild roster has been scanned leads to trouble.
		-- here's a crude gate for fast typists:
		if not self.hooked then
			self:LoadGuild()
		end
		LibStub("AceConfigDialog-3.0"):Open(nametag)
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(self, "oga", nametag, input)
	end
end


--[[
Observations:
1) There are a lot of funky ways to screw up text before AddMessage ever sees
   it, some from Blizzard and some by other addons.  This means that debugging
   the pattern matching after each patch can be a long and tedious process,
   made tractable by tons of debugging statements.
2) Testing for the same debug flag over and over on a hot code path is bad,
   especially when it will almost never change.  Debug code needs to either be
   there all the time or none of the time, when possible.
3) When faced with an odd pattern-not-matching situation, we *really* need to
   be able to turn on debugging without asking a user to install a specical
   version with extra print code.

So instead of writing a function, we write the source for that function and
surround the very slow parts with some sentinel tokens.  When starting up for
a game session, the tokens are stripped -- either removing the debugging code
or leaving it behind -- and the result is compiled then.  This is as close to
a macro processor as we can get without using a modified Lua.

Added fun from the fact that loadstring runs in the global env.  WTB Lua 5.2.
]]
function addon:HookChat ()
	self:UnhookChat()

	self.prat = self.prat or handle_prat()
	handle_prat = nil
	if self.prat == "bail" then
		return
	elseif self.prat == "cleanup" then
		self.rank_letter_map = nil
		self.alts_to_main = nil
		self.rank_lookup = nil
		return
	end

	state.channel_match_pattern = "|Hchannel:(.-)|h"
	if CHAT_TIMESTAMP_FORMAT == nil then
		state.channel_match_pattern = "^" .. state.channel_match_pattern
	end
	state.dprint("channel pattern:", state.channel_match_pattern)

	if not self.AddMessage then
		local func_def
		if self.db.profile.hcd_debug then
			state.hcd_debugging_frame = _G["ChatFrame" .. self.db.profile.hcd_debugging_frame]
			func_def = self.func_def:gsub("%-%-<HCD",""):gsub("%-%->HCD","")
		else
			func_def = self.func_def:gsub("%-%-<HCD.-%-%->HCD","")
		end
		local func,err = loadstring(func_def, "Ouro Guild Alts")
		if func then
			setfenv (func, state)
			self.AddMessage = func()   -- don't pcall this for now
			self.func_def = nil
		else
			self:Print("ERROR: Could not create chat hook functions, giving up.")
			return self:Disable()
		end
	end

	for i = 1, self.db.profile.max_chatframe_num do
		local cf = _G["ChatFrame"..i]
		if not cf then break end
		self:RawHook (cf, "AddMessage", --[[overridesecure=]]true)
	end
	self.hooked = true
	if state.debugging() then
		_G.OGA = self
	end
end

function addon:UnhookChat()
	self:UnhookAll()
	self.hooked = nil
end


function addon:LoadGuild ()
	--self:FriendsFrame_Pause()

	if not IsInGuild() then return end

	local nAlt   = 0
	local nGuild = GetNumGuildMembers(true) or 0

	-- Try and compensate for laggy-ass servers.
	-- http://xkcd.com/37/
	if nGuild == 0 then
		self.lagloop = self.lagloop and (self.lagloop - 1) or 5
		if self.lagloop > 0 then
			return GuildRoster()   -- dispatch a request
		else
			self:Print("ERROR: Repeatedly failed to retrieve guild roster, giving up.")
			return self:Disable()
		end
	end
	self.lagloop = nil
	self:UnregisterEvent("GUILD_ROSTER_SHOW")
	self:UnregisterEvent("GUILD_ROSTER_UPDATE")

	-- Build the map
	local GetGuildRosterInfo = _G.GetGuildRosterInfo
	self.alts_to_main = {}
	self:UpdatePattern (nil, self.db.profile.alt_pattern)  -- builds lua_pattern
	local altrank_pattern = '%f[%w]' .. altrank .. '%f[%W]'
	for i = 1, nGuild do
		local cleannote
		local name, rank, rankIndex, level, class, zone, note, officernote, online, status = GetGuildRosterInfo(i)

		if type(officernote) == 'string' and officernote ~= "" then
			cleannote = officernote:match(lua_pattern)
		end
		if (not cleannote) and type(note) == 'string' and note ~= "" then
			cleannote = note:match(lua_pattern)
		end
		if (not cleannote) and self.db.profile.use_alt_rank and rank:lower():match(altrank_pattern) then
			state.dprint("whoa, rank match!", rank)
			cleannote = note
		end
		if cleannote and cleannote ~= "" and name:lower() ~= cleannote:lower() then
			self.alts_to_main[name] = cleannote
			nAlt = nAlt + 1
		end
	end

	self:Printf("Found %d guild members, %d of whom are labeled as alts.", nGuild, nAlt)
	guildname = GetGuildInfo("player") or 'NONE'
	guildkey = ("<%s> - %s"):format(guildname, GetRealmName())
	isofficer = not not CanGuildInvite()
	self.rank_lookup = nil
	self.rank_letter_map = nil
	if self.PrepareRankTagging then
		self:PrepareRankTagging()
	end
	self:UpdateRankTagging()

	--self:FriendsFrame_Resume()
	if not self.hooked then
		self:HookChat()
	end
end


--[[
	As of 4.0, raw channel text looks like this:
		|Hchannel:6|h[6. foo]|h |Hplayer:Person:linenumber|h[Person]|h: hello
	and displays like
		[6. foo] [Person]: hello
	The channel 'numbers' of other channels are:  "GUILD" "o" "Party"/"party" "raid"
	Amusingly, "Party" is for normal members and "party" is for the leader.  The raid
	leader chat is already picked up by "raid".
	For system login messages, the ":linenumber" part is not present:
		|Hplayer:Person|h[Person]|h has come online.

	Colorized raid line looks like
	|Hchannel:RAID|h[Raid]|h |Hplayer:Me:linenumber:RAID|h[|cffxxxxxxMe|r]|h: hello

	If timestamps are in use, they're just tacked onto the beginning of the whole thing.
]]


-- Returns (1) the sender's name and (2) the pattern to replace in chat.  If
-- class colors are in use, (2) will be the colorized version of (1), else just
-- a copy of (1).
function state.extract_name_and_target (msg)
	local n, rest = msg:match("|Hplayer:(.-)[:|](.-)|h[: ]")
	if not n then state.dprint(state.author_debug, "n error on", msg:gsub("\124","\124\124")) return end
	if not rest then state.dprint(state.author_debug, "rest error on", (n or "<n nil>"):gsub("\124","\124\124"), "from", msg:gsub("\124","\124\124")) return end
	local c = rest:match("|h%[(|cff......)")
	if c then
		c = c .. n .. "|r"
		return n, c
	else return n, n end
end


addon.func_def = [=[
return function (self, frame, msg, ...)
	if not self.db.profile.enable then
		-- We shouldn't even be here today!
		self:Disable()
		return self.hooks[frame].AddMessage(frame,msg,...)
	end

	local channel = msg:match(channel_match_pattern)
	channel = channel and channel:lower()
	--<HCD
	dprint(hcd_debugging_frame, frame:GetName() .. "<" .. (channel or "") .. ">:  " .. (msg:gsub("\124", "\124\124")), ...)
	-->HCD

	if channel == "guild" or channel == "officer" or channel == "party" or channel == "raid"
	   --or (channel == nil and msg:find("|h has come online."))   -- triggers on Real ID logins also
	   or (channel == nil and msg:find("|Hplayer.*|h has come online."))
	then
		local sender, replace_target = extract_name_and_target(msg)

		if sender == me
			--<HCD
			and not debugging()
			-->HCD
		then
			return orig(frame,msg,...)
		end

		-- Try to look it up
		local mainname = self.alts_to_main[sender]
		local replacement

		if mainname then
			replacement = replace_target .. " (" .. mainname .. ")"
		end

		if do_rank_tags and (tag_channels[channel]
			                 or debugging()
							)
		then
			local r = self.rank_lookup[sender]
			dprint("rank lookup on "..sender.." found "..(r or "nil"))
			if r then
				replacement = (replacement or replace_target) .. ';' .. r
			end
		end

		if replacement then
			--<HCD
			dprint(hcd_debugging_frame, "target: "..(replace_target:gsub("\124","\124\124")))
			dprint(hcd_debugging_frame, "replacement: "..(replacement:gsub("\124","\124\124")))
			-->HCD
			msg = msg:gsub("|h%["..replace_target.."%]|h",
			               "|h%["..replacement.."%]|h")
		end

	end

	return orig(frame,msg,...)
end
]=]

-- vim:noet
